/*
Build: gcc -o fpfail.exe fpfail.c
*/

#include <stdio.h>
#include <stdlib.h>

const unsigned int N = 5;                               //vector length

double SumUpFPU(double *p,unsigned int m);              //protos
double SumUpXMM(double *p,unsigned int m);
                               
int main(int argc, char **argv)
{
  double V1[5] = {1e20,17.0,-10.0,130.0,-1e20};
  double V2[5] = {1e20,-10.0,130.0,-1e20,17.0};
  double V3[5] = {1e20,17.0,-1e20,-10.0,130.0};
  double V4[5] = {1e20,-10.0,-1e20,130.0,17.0};
  double V5[5] = {1e20,-1e20,17.0,-10.0,130.0};
  double V6[5] = {1e20,17.0,130.0,-1e20,-10.0};
  
  double sum1FPU,sum1XMM;
  double sum2FPU,sum2XMM;
  double sum3FPU,sum3XMM;
  double sum4FPU,sum4XMM;
  double sum5FPU,sum5XMM;
  double sum6FPU,sum6XMM;
  
  // addeded this
  __asm finit;

  sum1FPU = SumUpFPU(&V1[0],N);
  sum1XMM = SumUpXMM(&V1[0],N);
  sum2FPU = SumUpFPU(&V2[0],N);
  sum2XMM = SumUpXMM(&V2[0],N);
  sum3FPU = SumUpFPU(&V3[0],N);
  sum3XMM = SumUpXMM(&V3[0],N);
  sum4FPU = SumUpFPU(&V4[0],N);
  sum4XMM = SumUpXMM(&V4[0],N);
  sum5FPU = SumUpFPU(&V5[0],N);
  sum5XMM = SumUpXMM(&V5[0],N);
  sum6FPU = SumUpFPU(&V6[0],N);
  sum6XMM = SumUpXMM(&V6[0],N);

  
  printf("Sum 1 (FPU) = %.2lf Sum 1 (XMM) = %.2f\n",sum1FPU,sum1XMM);
  printf("Sum 2 (FPU) = %.2lf Sum 2 (XMM) = %.2f\n",sum2FPU,sum2XMM);
  printf("Sum 3 (FPU) = %.2lf Sum 3 (XMM) = %.2f\n",sum3FPU,sum3XMM);
  printf("Sum 4 (FPU) = %.2lf Sum 4 (XMM) = %.2f\n",sum4FPU,sum4XMM);
  printf("Sum 5 (FPU) = %.2lf Sum 5 (XMM) = %.2f\n",sum5FPU,sum5XMM);
  printf("Sum 6 (FPU) = %.2lf Sum 6 (XMM) = %.2f\n",sum6FPU,sum6XMM);
  printf("\nRight Sum   = 137.00\n");
  
  _getch();

  return 0;
}

/*
SumUpFPU
Purpose:        Sum up a double vector with the FPU.
Input:          [ebp+8]  = pointer to vector
                [ebp+12] = vector length
                [ebp-8]  = function result
Output:         vector sum (right or wrong)                
*/
__declspec(noinline,naked) double SumUpFPU(double *p,unsigned int m)
{
  __asm{
    mov        edx,[esp+4]//edx -> vector
    mov        ecx,[esp+8]//ecx = m (counter)
    fldz       //sum = 0
    b1:
    fadd       qword ptr [edx] //sum up
    lea        edx,[edx+8] //update pointer
    dec        ecx
    jnz        b1
    ret
    }
}

/*
SumUpXMM
Purpose:        Sum up a double vector with XMM registers.
Input:          [ebp+8]  = pointer to vector
                [ebp+12] = vector length
                [ebp-8]  = function result
Output:         vector sum (right or wrong)                
*/
__declspec(noinline,naked) double SumUpXMM(double *p,unsigned int m)
{
  __asm{
    push eax
    push eax
    mov        edx,[esp+12] //edx -> vector
    mov        ecx,[esp+16] //ecx = m (counter)
    pxor       xmm0,xmm0  //sum = 0
    b1: 
    addsd      xmm0,[edx]  //sum up
    lea        edx,[edx+8]//update pointer
    dec        ecx 
    jnz        b1
    movsd      [esp],xmm0//store result
    fld qword ptr [esp]
    pop ecx
    pop ecx
    ret
    }
}

